@@ -5,7 +5,7 @@ use strict;
 use warnings;
 use base qw( Exporter );
 
-our $VERSION = '0.05_01';
+our $VERSION = '0.07_01';
 
 use AnyEvent::RipeRedis::Error;
 
@@ -61,7 +61,7 @@ use constant {
   %ERROR_CODES,
 
   # Operation status
-  S_NEED_PERFORM => 1,
+  S_NEED_DO      => 1,
   S_IN_PROGRESS  => 2,
   S_DONE         => 3,
 
@@ -70,13 +70,13 @@ use constant {
   EOL_LENGTH => 2,
 };
 
-my %SUB_COMMANDS = (
+my %SUB_CMDS = (
   subscribe  => 1,
   psubscribe => 1,
 );
 
-my %SUBUNSUB_COMMANDS = (
-  %SUB_COMMANDS,
+my %SUBUNSUB_CMDS = (
+  %SUB_CMDS,
   unsubscribe  => 1,
   punsubscribe => 1,
 );
@@ -86,6 +86,20 @@ my %MESSAGE_TYPES = (
   pmessage => 1,
 );
 
+my %NEED_PREPROCESS = (
+  multi       => 1,
+  exec        => 1,
+  discard     => 1,
+  eval_cached => 1,
+  %SUBUNSUB_CMDS,
+);
+
+my %NEED_POSTPROCESS = (
+  info   => 1,
+  select => 1,
+  quit   => 1,
+);
+
 my %ERR_PREFS_MAP = (
   LOADING     => E_LOADING_DATASET,
   NOSCRIPT    => E_NO_SCRIPT,
@@ -120,11 +134,12 @@ sub new {
   $self->{password} = $params{password};
   $self->{database}
       = defined $params{database} ? $params{database} : D_DB_INDEX;
-  $self->{utf8}      = exists $params{utf8}      ? $params{utf8}      : 1;
-  $self->{reconnect} = exists $params{reconnect} ? $params{reconnect} : 1;
-  $self->{handle_params}    = $params{handle_params} || {};
-  $self->{on_connect}       = $params{on_connect};
-  $self->{on_disconnect}    = $params{on_disconnect};
+  $self->{utf8}          = exists $params{utf8} ? $params{utf8} : 1;
+  $self->{lazy}          = $params{lazy};
+  $self->{reconnect}     = exists $params{reconnect} ? $params{reconnect} : 1;
+  $self->{handle_params} = $params{handle_params} || {};
+  $self->{on_connect}    = $params{on_connect};
+  $self->{on_disconnect} = $params{on_disconnect};
   $self->{on_connect_error} = $params{on_connect_error};
 
   $self->connection_timeout( $params{connection_timeout} );
@@ -132,168 +147,26 @@ sub new {
   $self->min_reconnect_interval( $params{min_reconnect_interval} );
   $self->on_error( $params{on_error} );
 
-  $self->{_handle}           = undef;
-  $self->{_connected}        = 0;
-  $self->{_lazy_conn_state}  = $params{lazy};
-  $self->{_auth_state}       = S_NEED_PERFORM;
-  $self->{_select_db_state}  = S_NEED_PERFORM;
-  $self->{_ready}            = 0;
+  $self->_reset_internals;
   $self->{_input_queue}      = [];
   $self->{_temp_queue}       = [];
   $self->{_processing_queue} = [];
-  $self->{_txn_lock}         = 0;
   $self->{_channels}         = {};
   $self->{_channel_cnt}      = 0;
   $self->{_pchannel_cnt}     = 0;
-  $self->{_reconnect_timer}  = undef;
 
-  unless ( $self->{_lazy_conn_state} ) {
+  unless ( $self->{lazy} ) {
     $self->_connect;
   }
 
   return $self;
 }
 
-sub info {
-  my $self = shift;
-  my $cmd  = $self->_prepare( 'info', [@_] );
-
-  weaken($self);
-  my $on_reply = $cmd->{on_reply};
-
-  $cmd->{on_reply} = sub {
-    my $reply = shift;
-    my $err   = shift;
-
-    if ( defined $err ) {
-      $on_reply->( undef, $err );
-      return;
-    }
-
-    $reply = { map { split( m/:/, $_, 2 ) }
-        grep { m/^[^#]/ } split( EOL, $reply ) };
-
-    $on_reply->($reply);
-  };
-
-  $self->_execute($cmd);
-
-  return;
-}
-
-sub select {
-  my $self = shift;
-  my $cmd  = $self->_prepare( 'select', [@_] );
-
-  weaken($self);
-
-  my $on_reply = $cmd->{on_reply};
-  my $database = $cmd->{args}[0];
-
-  $cmd->{on_reply} = sub {
-    my $reply = shift;
-    my $err   = shift;
-
-    if ( defined $err ) {
-      $on_reply->( undef, $err );
-      return;
-    }
-
-    $self->{database} = $database;
-
-    $on_reply->($reply);
-  };
-
-  $self->_execute($cmd);
-
-  return;
-}
-
-sub multi {
-  my $self = shift;
-  my $cmd  = $self->_prepare( 'multi', [@_] );
-
-  $self->{_txn_lock} = 1;
-  $self->_execute($cmd);
-
-  return;
-}
-
-sub exec {
-  my $self = shift;
-  my $cmd  = $self->_prepare( 'exec', [@_] );
-
-  $self->{_txn_lock} = 0;
-  $self->_execute($cmd);
-
-  return;
-}
-
-sub eval_cached {
-  my $self = shift;
-  my $cmd  = $self->_prepare( 'evalsha', [@_] );
-
-  my $script = $cmd->{args}[0];
-  unless ( exists $EVAL_CACHE{$script} ) {
-    $EVAL_CACHE{$script} = sha1_hex($script);
-  }
-  $cmd->{args}[0] = $EVAL_CACHE{$script};
-
-  {
-    weaken($self);
-    weaken( my $cmd = $cmd );
-
-    my $on_reply = $cmd->{on_reply};
-
-    $cmd->{on_reply} = sub {
-      my $reply = shift;
-      my $err   = shift;
-
-      if ( defined $err ) {
-        if ( $err->code == E_NO_SCRIPT ) {
-          $cmd->{kwd}     = 'eval';
-          $cmd->{args}[0] = $script;
-
-          $self->_push_write($cmd);
-
-          return;
-        }
-
-        $on_reply->( $reply, $err );
-
-        return;
-      }
-
-      $on_reply->($reply);
-    };
-  }
-
-  $self->_execute($cmd);
-
-  return;
-}
-
-sub quit {
-  my $self = shift;
-  my $cmd  = $self->_prepare( 'quit', [@_] );
-
-  weaken($self);
-  my $on_reply = $cmd->{on_reply};
-
-  $cmd->{on_reply} = sub {
-    my $reply = shift;
-    my $err   = shift;
-
-    if ( defined $err ) {
-      $on_reply->( undef, $err );
-      return;
-    }
-
-    $self->_disconnect;
-
-    $on_reply->($reply);
-  };
+sub execute {
+  my $self     = shift;
+  my $cmd_name = shift;
 
+  my $cmd = $self->_prepare( $cmd_name, [@_] );
   $self->_execute($cmd);
 
   return;
@@ -307,63 +180,6 @@ sub disconnect {
   return;
 }
 
-# Generate sub/unsub methods
-{
-  no strict 'refs';
-
-  foreach my $kwd ( keys %SUBUNSUB_COMMANDS ) {
-    *{$kwd} = sub {
-      my $self = shift;
-      my $cmd = $self->_prepare( $kwd, [@_] );
-
-      if ( exists $SUB_COMMANDS{ $cmd->{kwd} }
-        && !defined $cmd->{on_message} )
-      {
-        croak q{"on_message" callback must be specified};
-      }
-
-      if ( $self->{_txn_lock} ) {
-        AE::postpone(
-          sub {
-            my $err = _new_error(
-              qq{Command "$cmd->{kwd}" not allowed after "multi" command.}
-                  . ' First, the transaction must be finalized.',
-              E_OPRN_ERROR
-            );
-
-            $cmd->{on_reply}->( undef, $err );
-          }
-        );
-
-        return;
-      }
-
-      if ( @{ $cmd->{args} } ) {
-        $cmd->{reply_cnt} = scalar @{ $cmd->{args} };
-      }
-
-      $self->_execute($cmd);
-
-      return;
-    };
-  }
-}
-
-sub host {
-  my $self = shift;
-  return $self->{host};
-}
-
-sub port {
-  my $self = shift;
-  return $self->{port};
-}
-
-sub selected_database {
-  my $self = shift;
-  return $self->{database};
-}
-
 sub on_error {
   my $self = shift;
 
@@ -388,6 +204,13 @@ sub on_error {
 {
   no strict 'refs';
 
+  foreach my $name ( qw( host port database ) ) {
+    *{$name} = sub {
+      my $self = shift;
+      return $self->{$name};
+    }
+  }
+
   foreach my $name ( qw( connection_timeout read_timeout
       min_reconnect_interval ) )
   {
@@ -468,13 +291,13 @@ sub _create_on_connect {
       $self->{_auth_state} = S_DONE;
     }
     if ( $self->{database} == 0 ) {
-      $self->{_select_db_state} = S_DONE;
+      $self->{_db_selection_state} = S_DONE;
     }
 
-    if ( $self->{_auth_state} == S_NEED_PERFORM ) {
+    if ( $self->{_auth_state} == S_NEED_DO ) {
       $self->_auth;
     }
-    elsif ( $self->{_select_db_state} == S_NEED_PERFORM ) {
+    elsif ( $self->{_db_selection_state} == S_NEED_DO ) {
       $self->_select_database;
     }
     else {
@@ -500,7 +323,6 @@ sub _create_on_connect_error {
       "Can't connect to $self->{host}:$self->{port}: $err_msg",
       E_CANT_CONN
     );
-
     $self->_disconnect($err);
   };
 }
@@ -513,7 +335,6 @@ sub _create_on_rtimeout {
   return sub {
     if ( @{ $self->{_processing_queue} } ) {
       my $err = _new_error( 'Read timed out.', E_READ_TIMEDOUT );
-
       $self->_disconnect($err);
     }
     else {
@@ -530,7 +351,6 @@ sub _create_on_eof {
   return sub {
     my $err = _new_error( 'Connection closed by remote host.',
         E_CONN_CLOSED_BY_REMOTE_HOST );
-
     $self->_disconnect($err);
   };
 }
@@ -631,7 +451,6 @@ sub _create_on_read {
           else {
             my $err = _new_error( 'Unexpected reply received.',
                 E_UNEXPECTED_DATA );
-
             $self->_disconnect($err);
 
             return;
@@ -666,27 +485,37 @@ sub _create_on_read {
 }
 
 sub _prepare {
-  my $self = shift;
-  my $kwd  = shift;
-  my $args = shift;
+  my $self     = shift;
+  my $cmd_name = shift;
+  my $args     = shift;
 
-  my $cmd;
+  my $cbs;
   if ( ref( $args->[-1] ) eq 'HASH' ) {
-    $cmd = pop @{$args};
+    $cbs = pop @{$args};
   }
   else {
-    $cmd = {};
+    $cbs = {};
     if ( ref( $args->[-1] ) eq 'CODE' ) {
-      if ( exists $SUB_COMMANDS{$kwd} ) {
-        $cmd->{on_message} = pop @{$args};
+      if ( exists $SUB_CMDS{$cmd_name} ) {
+        $cbs->{on_message} = pop @{$args};
       }
       else {
-        $cmd->{on_reply} = pop @{$args};
+        $cbs->{on_reply} = pop @{$args};
       }
     }
   }
-  $cmd->{kwd}  = $kwd;
-  $cmd->{args} = $args;
+
+  my @kwds
+      = $cmd_name eq 'eval_cached'
+      ? ('evalsha')
+      : split( m/_/, lc($cmd_name) );
+
+  my $cmd = {
+    name => $cmd_name,
+    kwds => \@kwds,
+    args => $args,
+    %{$cbs},
+  };
 
   unless ( defined $cmd->{on_reply} ) {
     weaken($self);
@@ -709,21 +538,64 @@ sub _execute {
   my $self = shift;
   my $cmd  = shift;
 
+  if ( exists $NEED_PREPROCESS{ $cmd->{name} } ) {
+    if ( $cmd->{name} eq 'multi' ) {
+      $self->{_multi_mode} = 1;
+    }
+    elsif ( $cmd->{name} eq 'exec'
+      || $cmd->{name} eq 'discard' )
+    {
+      $self->{_multi_mode} = 0;
+    }
+    elsif ( $cmd->{name} eq 'eval_cached' ) {
+      my $script = $cmd->{args}[0];
+      unless ( exists $EVAL_CACHE{$script} ) {
+        $EVAL_CACHE{$script} = sha1_hex($script);
+      }
+      $cmd->{args}[0] = $EVAL_CACHE{$script};
+      $cmd->{script}  = $script;
+    }
+    else {    # subscribe, unsubscribe, psubscribe, punsubscribe
+      if ( exists $SUB_CMDS{ $cmd->{name} }
+        && !defined $cmd->{on_message} )
+      {
+        croak q{"on_message" callback must be specified};
+      }
+
+      if ( $self->{_multi_mode} ) {
+        AE::postpone {
+          my $err = _new_error(
+            qq{Command "$cmd->{name}" not allowed after "multi" command.}
+                . ' First, the transaction must be finalized.',
+            E_OPRN_ERROR
+          );
+          $cmd->{on_reply}->( undef, $err );
+        };
+
+        return;
+      }
+
+      if ( @{ $cmd->{args} } ) {
+        $cmd->{reply_cnt} = scalar @{ $cmd->{args} };
+      }
+    }
+  }
+
   unless ( $self->{_ready} ) {
     if ( defined $self->{_handle} ) {
       if ( $self->{_connected} ) {
         if ( $self->{_auth_state} == S_DONE ) {
-          if ( $self->{_select_db_state} == S_NEED_PERFORM ) {
+          if ( $self->{_db_selection_state} == S_NEED_DO ) {
             $self->_select_database;
           }
         }
-        elsif ( $self->{_auth_state} == S_NEED_PERFORM ) {
+        elsif ( $self->{_auth_state} == S_NEED_DO ) {
           $self->_auth;
         }
       }
     }
-    elsif ( $self->{_lazy_conn_state} ) {
-      $self->{_lazy_conn_state} = 0;
+    elsif ( $self->{lazy} ) {
+      undef $self->{lazy};
       $self->_connect;
     }
     elsif ( $self->{reconnect} ) {
@@ -745,16 +617,11 @@ sub _execute {
       }
     }
     else {
-      AE::postpone(
-        sub {
-          my $err = _new_error(
-            qq{Operation "$cmd->{kwd}" aborted: No connection to the server.},
-            E_NO_CONN
-          );
-
-          $cmd->{on_reply}->( undef, $err );
-        }
-      );
+      AE::postpone {
+        my $err = _new_error( qq{Operation "$cmd->{name}" aborted:}
+            . ' No connection to the server.', E_NO_CONN );
+        $cmd->{on_reply}->( undef, $err );
+      };
 
       return;
     }
@@ -774,7 +641,8 @@ sub _push_write {
   my $cmd  = shift;
 
   my $cmd_str = '';
-  foreach my $token ( $cmd->{kwd}, @{ $cmd->{args} } ) {
+  my @tokens  = ( @{ $cmd->{kwds} }, @{ $cmd->{args} } );
+  foreach my $token (@tokens) {
     unless ( defined $token ) {
       $token = '';
     }
@@ -783,11 +651,13 @@ sub _push_write {
     }
     $cmd_str .= '$' . length($token) . EOL . $token . EOL;
   }
-  $cmd_str = '*' . ( scalar( @{ $cmd->{args} } ) + 1 ) . EOL . $cmd_str;
+  $cmd_str = '*' . scalar(@tokens) . EOL . $cmd_str;
 
   my $handle = $self->{_handle};
 
-  if ( defined $self->{read_timeout} && !@{ $self->{_processing_queue} } ) {
+  if ( defined $self->{read_timeout}
+    && !@{ $self->{_processing_queue} } )
+  {
     $handle->rtimeout_reset;
     $handle->rtimeout( $self->{read_timeout} );
   }
@@ -805,14 +675,15 @@ sub _auth {
   $self->{_auth_state} = S_IN_PROGRESS;
 
   $self->_push_write(
-    { kwd  => 'auth',
+    { name => 'auth',
+      kwds => ['auth'],
       args => [ $self->{password} ],
 
       on_reply => sub {
         my $err = $_[1];
 
         if ( defined $err ) {
-          $self->{_auth_state} = S_NEED_PERFORM;
+          $self->{_auth_state} = S_NEED_DO;
           $self->_abort($err);
 
           return;
@@ -820,7 +691,7 @@ sub _auth {
 
         $self->{_auth_state} = S_DONE;
 
-        if ( $self->{_select_db_state} == S_NEED_PERFORM ) {
+        if ( $self->{_db_selection_state} == S_NEED_DO ) {
           $self->_select_database;
         }
         else {
@@ -838,25 +709,26 @@ sub _select_database {
   my $self = shift;
 
   weaken($self);
-  $self->{_select_db_state} = S_IN_PROGRESS;
+  $self->{_db_selection_state} = S_IN_PROGRESS;
 
   $self->_push_write(
-    { kwd  => 'select',
+    { name => 'select',
+      kwds => ['select'],
       args => [ $self->{database} ],
 
       on_reply => sub {
         my $err = $_[1];
 
         if ( defined $err ) {
-          $self->{_select_db_state} = S_NEED_PERFORM;
+          $self->{_db_selection_state} = S_NEED_DO;
           $self->_abort($err);
 
           return;
         }
 
-        $self->{_select_db_state} = S_DONE;
-        $self->{_ready}           = 1;
+        $self->{_db_selection_state} = S_DONE;
 
+        $self->{_ready} = 1;
         $self->_process_input_queue;
       },
     }
@@ -910,16 +782,25 @@ sub _process_error {
       q{Don't know how process error message. Processing queue is empty.},
       E_UNEXPECTED_DATA
     );
-
     $self->_disconnect($err);
 
     return;
   }
 
+  if ( $cmd->{name} eq 'eval_cached'
+    && $err_code == E_NO_SCRIPT )
+  {
+    $cmd->{kwds}[0] = 'eval';
+    $cmd->{args}[0] = $cmd->{script};
+
+    $self->_push_write($cmd);
+
+    return;
+  }
+
   if ( ref($reply) ) {
     my $err = _new_error(
-        qq{Operation "$cmd->{kwd}" completed with errors.}, $err_code );
-
+        qq{Operation "$cmd->{name}" completed with errors.}, $err_code );
     $cmd->{on_reply}->( $reply, $err );
   }
   else {
@@ -942,7 +823,6 @@ sub _process_message {
           . qq{ Unknown channel or pattern "$msg->[1]".},
       E_UNEXPECTED_DATA
     );
-
     $self->_disconnect($err);
 
     return;
@@ -968,22 +848,21 @@ sub _process_success {
       q{Don't know how process reply. Processing queue is empty.},
       E_UNEXPECTED_DATA
     );
-
     $self->_disconnect($err);
 
     return;
   }
 
-  if ( exists $SUBUNSUB_COMMANDS{ $cmd->{kwd} } ) {
-    if ( $cmd->{kwd} eq 'subscribe' ) {
+  if ( exists $SUBUNSUB_CMDS{ $cmd->{name} } ) {
+    if ( $cmd->{name} eq 'subscribe' ) {
       $self->{_channels}{ $reply->[1] } = $cmd;
       $self->{_channel_cnt}++;
     }
-    elsif ( $cmd->{kwd} eq 'psubscribe' ) {
+    elsif ( $cmd->{name} eq 'psubscribe' ) {
       $self->{_channels}{ $reply->[1] } = $cmd;
       $self->{_pchannel_cnt}++;
     }
-    elsif ( $cmd->{kwd} eq 'unsubscribe' ) {
+    elsif ( $cmd->{name} eq 'unsubscribe' ) {
       unless ( defined $cmd->{reply_cnt} ) {
         $cmd->{reply_cnt} = $self->{_channel_cnt};
       }
@@ -1003,8 +882,24 @@ sub _process_success {
     $reply = $reply->[2];
   }
 
-  if ( !defined $cmd->{reply_cnt} || --$cmd->{reply_cnt} == 0 ) {
+  if ( !defined $cmd->{reply_cnt}
+    || --$cmd->{reply_cnt} == 0 )
+  {
     shift @{ $self->{_processing_queue} };
+
+    if ( exists $NEED_POSTPROCESS{ $cmd->{name} } ) {
+      if ( $cmd->{name} eq 'info' ) {
+        $reply = { map { split( m/:/, $_, 2 ) }
+            grep { m/^[^#]/ } split( EOL, $reply ) };
+      }
+      elsif ( $cmd->{name} eq 'select' ) {
+        $self->{database} = $cmd->{args}[0];
+      }
+      else {    # quit
+        $self->_disconnect;
+      }
+    }
+
     $cmd->{on_reply}->($reply);
   }
 
@@ -1019,14 +914,8 @@ sub _disconnect {
 
   if ( defined $self->{_handle} ) {
     $self->{_handle}->destroy;
-    undef $self->{_handle};
   }
-  $self->{_connected}       = 0;
-  $self->{_auth_state}      = S_NEED_PERFORM;
-  $self->{_select_db_state} = S_NEED_PERFORM;
-  $self->{_ready}           = 0;
-  $self->{_txn_lock}        = 0;
-
+  $self->_reset_internals;
   $self->_abort($err);
 
   if ( $was_connected && defined $self->{on_disconnect} ) {
@@ -1036,6 +925,20 @@ sub _disconnect {
   return;
 }
 
+sub _reset_internals {
+  my $self = shift;
+
+  $self->{_handle}             = undef;
+  $self->{_connected}          = 0;
+  $self->{_auth_state}         = S_NEED_DO;
+  $self->{_db_selection_state} = S_NEED_DO;
+  $self->{_ready}              = 0;
+  $self->{_multi_mode}         = 0;
+  $self->{_reconnect_timer}    = undef;
+
+  return;
+}
+
 sub _abort {
   my $self = shift;
   my $err  = shift;
@@ -1079,9 +982,8 @@ sub _abort {
     }
 
     foreach my $cmd (@queued_commands) {
-      my $err = _new_error( qq{Operation "$cmd->{kwd}" aborted: $err_msg},
+      my $err = _new_error( qq{Operation "$cmd->{name}" aborted: $err_msg},
           $err_code );
-
       $cmd->{on_reply}->( undef, $err );
     }
   }
@@ -1105,14 +1007,13 @@ sub _new_error {
 
 sub AUTOLOAD {
   our $AUTOLOAD;
-  my $method = $AUTOLOAD;
-  $method =~ s/^.+:://;
-  my ( $kwd, @extra_args ) = split( m/_/, lc($method) );
+  my $cmd_name = $AUTOLOAD;
+  $cmd_name =~ s/^.+:://;
 
   my $sub = sub {
     my $self = shift;
-    my $cmd  = $self->_prepare( $kwd, [ @extra_args, @_ ] );
 
+    my $cmd = $self->_prepare( $cmd_name, [@_] );
     $self->_execute($cmd);
 
     return;
@@ -1120,7 +1021,7 @@ sub AUTOLOAD {
 
   do {
     no strict 'refs';
-    *{$method} = $sub;
+    *{$cmd_name} = $sub;
   };
 
   goto &{$sub};
@@ -1133,7 +1034,7 @@ sub DESTROY {
     my @queued_commands = $self->_queued_commands;
 
     foreach my $cmd (@queued_commands) {
-      warn qq{Operation "$cmd->{kwd}" aborted:}
+      warn qq{Operation "$cmd->{name}" aborted:}
           . " Client object destroyed prematurely.\n";
     }
   }
@@ -1294,7 +1195,7 @@ TRUE (default), the client try to restore the connection when you execute next
 command. The client try to reconnect only once and, if attempt fails, the error
 object is passed to command callback. If you need several attempts of the
 reconnection, you must retry a command from the callback as many times, as you
-need. Such behavior allows to manage reconnection procedure.
+need. Such behavior allows to control reconnection procedure.
 
 Enabled by default.
 
@@ -1416,6 +1317,26 @@ You can execute multi-word commands like this:
     }
   );
 
+=head2 execute( $command, [ @args ] [, $cb->( $reply, $err ) ] )
+
+An alternative method to execute commands. In some cases it can be more
+convenient.
+
+  $redis->execute( 'get', 'foo',
+    sub {
+      my $reply = shift;
+      my $err   = shift;
+
+      if ( defined $err ) {
+        # error handling...
+
+        return;
+      }
+
+      print "$reply\n";
+    }
+  );
+
 =head1 TRANSACTIONS
 
 The detailed information about the Redis transactions can be found here:
@@ -1887,7 +1808,7 @@ Get current port of the client.
 
 Selects the database by numeric index.
 
-=head2 selected_database()
+=head2 database()
 
 Get selected database index.
 
